# -*- coding: utf-8 -*-
import unittest
from unittest.mock import MagicMock, patch, call

from ncclient.transport.ssh import SSHSession
from ncclient.transport import AuthenticationError, SessionCloseError, NetconfBase
import paramiko
from ncclient.devices.junos import JunosDeviceHandler

try:
    import selectors
except ImportError:
    import selectors2 as selectors


SSH_KEYTYPE = 'ssh-ed25519'
SSH_PUBKEY = 'test/unit/transport/certs/id_ed25519_test.pub'
SSH_CERT = 'test/unit/transport/certs/id_ed25519_test-cert.pub'

reply_data = """<rpc-reply xmlns:junos="http://xml.juniper.net/junos/12.1X46/junos" attrib1 = "test">
    <software-information>
        <host-name>R1</host-name>
        <product-model>firefly-perimeter</product-model>
        <product-name>firefly-perimeter</product-name>
        <package-information>
            <name>junos</name>
            <comment>JUNOS Software Release [12.1X46-D10.2]</comment>
        </package-information>
    </software-information>
    <cli>
        <banner></banner>
    </cli>
</rpc-reply>"""

reply_ok = """<rpc-reply>
    <ok/>
<rpc-reply/>"""

# A buffer of data with two complete messages and an incomplete message
rpc_reply = reply_data + "\n]]>]]>\n" + reply_ok + "\n]]>]]>\n" + reply_ok

reply_ok_chunk = "\n#%d\n%s\n##\n" % (len(reply_ok), reply_ok)

# einarnn: this test message had to be reduced in size as the improved
# 1.1 parsing finds a whole fragment in it, so needed to have less
# data in it than the terminating '>'
reply_ok_partial_chunk = "\n#%d\n%s" % (len(reply_ok), reply_ok[:-1])

# A buffer of data with two complete messages and an incomplete message
rpc_reply11 = "\n#%d\n%s\n#%d\n%s\n##\n%s%s" % (
    30, reply_data[:30], len(reply_data[30:]), reply_data[30:],
    reply_ok_chunk, reply_ok_partial_chunk)


rpc_reply_part_1 = """<rpc-reply xmlns:junos="http://xml.juniper.net/junos/12.1X46/junos" attrib1 = "test">
    <software-information>
        <host-name>R1</host-name>
        <product-model>firefly-perimeter</product-model>
        <product-name>firefly-perimeter</product-name>
        <package-information>
            <name>junos</name>
            <comment>JUNOS Software Release [12.1X46-D10.2]</comment>
        </package-information>
    </software-information>
    <cli>
        <banner></banner>
    </cli>
</rpc-reply>
]]>]]"""

rpc_reply_part_2 = """>
<rpc-reply>
    <ok/>
<rpc-reply/>"""


class TestSSH(unittest.TestCase):

    def _test_parsemethod(self, mock_dispatch, parsemethod, reply, ok_chunk,
                          expected_messages):
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._buffer.write(bytes(reply, "utf-8"))
        remainder = bytes(ok_chunk, "utf-8")
        parsemethod(obj)

        for i in range(0, len(expected_messages)):
            call = mock_dispatch.call_args_list[i][0][0]
            self.assertEqual(call, expected_messages[i])

        self.assertEqual(obj._buffer.getvalue(), remainder)

    @patch('ncclient.transport.ssh.Session._dispatch_message')
    def test_parse(self, mock_dispatch):
        self._test_parsemethod(mock_dispatch, SSHSession._parse, rpc_reply,
                               "\n" + reply_ok, [reply_data])

    @patch('ncclient.transport.ssh.Session._dispatch_message')
    def test_parse11(self, mock_dispatch):
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._buffer.write(bytes(rpc_reply11, "utf-8"))
        remainder = bytes(reply_ok_partial_chunk, "utf-8")
        obj.parser._parse11()

        expected_messages = [reply_data, reply_ok]
        for i in range(0, len(expected_messages)):
            call = mock_dispatch.call_args_list[i][0][0]
            self.assertEqual(call, expected_messages[i])

        self.assertEqual(obj._buffer.getvalue(), remainder)

    @patch('ncclient.transport.ssh.Session._dispatch_message')
    def test_parse_incomplete_delimiter(self, mock_dispatch):
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        b = bytes(rpc_reply_part_1, "utf-8")
        obj._buffer.write(b)
        obj._parse()
        self.assertFalse(mock_dispatch.called)
        b = bytes(rpc_reply_part_2, "utf-8")
        obj._buffer.write(b)
        obj._parse()
        self.assertTrue(mock_dispatch.called)

    @patch('paramiko.transport.Transport.auth_publickey')
    @patch('paramiko.agent.AgentSSH.get_keys')
    def test_auth_agent(self, mock_get_key, mock_auth_public_key):
        key = paramiko.PKey(msg="hello")
        mock_get_key.return_value = [key]
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._transport = paramiko.Transport(MagicMock())
        obj._auth('user', 'password', [], True, True)
        self.assertEqual(
            (mock_auth_public_key.call_args_list[0][0][1]).__repr__(),
            key.__repr__())

    @patch('paramiko.transport.Transport.auth_publickey')
    @patch('paramiko.agent.AgentSSH.get_keys')
    def test_auth_agent_with_key(self, mock_get_key, mock_auth_public_key):
        expected_blob = paramiko.PublicBlob.from_file(SSH_PUBKEY).key_blob
        expected = paramiko.PKey.from_type_string(SSH_KEYTYPE, expected_blob)
        expected.load_certificate(SSH_PUBKEY)

        agentkey_blob = paramiko.PublicBlob.from_file(SSH_PUBKEY).key_blob
        agentkey = paramiko.PKey.from_type_string(SSH_KEYTYPE, agentkey_blob)
        mock_get_key.return_value = [agentkey]

        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._transport = paramiko.Transport(MagicMock())
        obj._auth('user', 'password', [SSH_PUBKEY], True, True)
        self.assertEqual(
            (mock_auth_public_key.call_args_list[0][0][1]).__repr__(),
            expected.__repr__())
        self.assertEqual(
            (mock_auth_public_key.call_args_list[0][0][1]).public_blob,
            expected.public_blob)

    @patch('paramiko.transport.Transport.auth_publickey')
    @patch('paramiko.agent.AgentSSH.get_keys')
    def test_auth_agent_with_cert(self, mock_get_key, mock_auth_public_key):
        expected_blob = paramiko.PublicBlob.from_file(SSH_PUBKEY).key_blob
        expected = paramiko.PKey.from_type_string(SSH_KEYTYPE, expected_blob)
        expected.load_certificate(SSH_CERT)

        agentkey_blob = paramiko.PublicBlob.from_file(SSH_PUBKEY).key_blob
        agentkey = paramiko.PKey.from_type_string(SSH_KEYTYPE, agentkey_blob)
        mock_get_key.return_value = [agentkey]

        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._transport = paramiko.Transport(MagicMock())
        obj._auth('user', 'password', [SSH_CERT], True, True)
        self.assertEqual(
            (mock_auth_public_key.call_args_list[0][0][1]).__repr__(),
            expected.__repr__())
        self.assertEqual(
            (mock_auth_public_key.call_args_list[0][0][1]).public_blob,
            expected.public_blob)

    @patch('paramiko.transport.Transport.auth_publickey')
    @patch('paramiko.agent.AgentSSH.get_keys')
    def test_auth_agent_exception(self, mock_get_key, mock_auth_public_key):
        key = paramiko.PKey()
        mock_get_key.return_value = [key]
        mock_auth_public_key.side_effect = paramiko.ssh_exception.AuthenticationException
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._transport = paramiko.Transport(MagicMock())
        self.assertRaises(AuthenticationError,
            obj._auth,'user', None, [], True, False)

    @patch('paramiko.transport.Transport.auth_publickey')
    @patch('paramiko.PKey.from_path')
    def test_auth_keyfiles(self, mock_get_key, mock_auth_public_key):
        key = paramiko.PKey()
        mock_get_key.return_value = key
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._transport = paramiko.Transport(MagicMock())
        obj._auth('user', 'password', ["key_file_name"], False, True)
        self.assertEqual(
            (mock_auth_public_key.call_args_list[0][0][1]).__repr__(),
            key.__repr__())

    @patch('paramiko.transport.Transport.auth_publickey')
    @patch('paramiko.PKey.from_path')
    def test_auth_keyfiles_exception(self, mock_get_key, mock_auth_public_key):
        key = paramiko.PKey()
        mock_get_key.side_effect = paramiko.ssh_exception.PasswordRequiredException
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._transport = paramiko.Transport(MagicMock())
        self.assertRaises(AuthenticationError,
            obj._auth,'user', None, ["key_file_name"], False, True)

    @patch('os.path.isfile')
    @patch('paramiko.transport.Transport.auth_publickey')
    @patch('paramiko.PKey.from_path')
    def test_auth_default_keyfiles(self, mock_get_key, mock_auth_public_key,
                                   mock_is_file):
        key = paramiko.PKey()
        mock_get_key.return_value = key
        mock_is_file.return_value = True
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._transport = paramiko.Transport(MagicMock())
        obj._auth('user', 'password', [], False, True)
        self.assertEqual(
            (mock_auth_public_key.call_args_list[0][0][1]).__repr__(),
            key.__repr__())

    @patch('os.path.isfile')
    @patch('paramiko.transport.Transport.auth_publickey')
    @patch('paramiko.PKey.from_path')
    def test_auth_default_keyfiles_exception(self, mock_get_key,
                                             mock_auth_public_key, mock_is_file):
        key = paramiko.PKey()
        mock_is_file.return_value = True
        mock_get_key.side_effect = paramiko.ssh_exception.PasswordRequiredException
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._transport = paramiko.Transport(MagicMock())
        self.assertRaises(AuthenticationError,
			              obj._auth,'user', None, [], False, True)

    @patch('paramiko.transport.Transport.auth_password')
    def test_auth_password(self, mock_auth_password):
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._transport = paramiko.Transport(MagicMock())
        obj._auth('user', 'password', [], False, True)
        self.assertEqual(
            mock_auth_password.call_args_list[0][0],
            ('user',
             'password'))

    @patch('paramiko.transport.Transport.auth_password')
    def test_auth_exception(self, mock_auth_password):
        mock_auth_password.side_effect = Exception
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._transport = paramiko.Transport(MagicMock())
        self.assertRaises(AuthenticationError,
            obj._auth, 'user', 'password', [], False, True)

    def test_auth_no_methods_exception(self):
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._transport = paramiko.Transport(MagicMock())
        self.assertRaises(AuthenticationError,
            obj._auth,'user', None, [], False, False)

    @patch('paramiko.transport.Transport.close')
    def test_close(self, mock_close):
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._transport = paramiko.Transport(MagicMock())
        obj._transport.active = True
        obj._connected = True
        obj.close()
        mock_close.assert_called_once_with()
        self.assertFalse(obj._connected)

    @patch('paramiko.hostkeys.HostKeys.load')
    def test_load_host_key(self, mock_load):
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj.load_known_hosts("file_name")
        mock_load.assert_called_once_with("file_name")

    @patch('os.path.expanduser')
    @patch('paramiko.hostkeys.HostKeys.load')
    def test_load_host_key_2(self, mock_load, mock_os):
        mock_os.return_value = "file_name"
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj.load_known_hosts()
        mock_load.assert_called_once_with("file_name")

    @patch('os.path.expanduser')
    @patch('paramiko.hostkeys.HostKeys.load')
    def test_load_host_key_IOError(self, mock_load, mock_os):
        mock_os.return_value = "file_name"
        mock_load.side_effect = IOError
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj.load_known_hosts()
        mock_load.assert_called_with("file_name")

    @patch('ncclient.transport.ssh.hexlify')
    @patch('os.path.expanduser')
    @patch('paramiko.HostKeys')
    @patch('paramiko.Transport')
    @patch('ncclient.transport.ssh.SSHSession._post_connect')
    @patch('ncclient.transport.ssh.SSHSession._auth')
    def test_ssh_known_hosts(self, mock_auth, mock_pc,
                             mock_transport, mock_hk,
                             mock_os, mock_hex):
        mock_os.return_value = "file_name"
        hk_inst = MagicMock(check=MagicMock(return_value=True))
        mock_hk.return_value = hk_inst
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj.connect(host='h', sock=MagicMock())
        hk_inst.load.assert_called_once_with('file_name')
        mock_os.assert_called_once_with('~/.ssh/known_hosts')

    @patch('ncclient.transport.ssh.hexlify')
    @patch('ncclient.transport.ssh.open')
    @patch('os.path.expanduser')
    @patch('paramiko.HostKeys')
    @patch('paramiko.Transport')
    @patch('paramiko.SSHConfig')
    @patch('ncclient.transport.ssh.SSHSession._post_connect')
    @patch('ncclient.transport.ssh.SSHSession._auth')
    def test_ssh_known_hosts_2(self, mock_auth, mock_pc,
                               mock_sshc, mock_transport, mock_hk,
                               mock_os, mock_open, mock_hex):
        mock_os.return_value = "file_name"
        hk_inst = MagicMock(check=MagicMock(return_value=True))
        mock_hk.return_value = hk_inst
        config = {'userknownhostsfile': 'known_hosts_file'}
        mock_sshc.return_value = MagicMock(lookup=lambda _h: config)
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj.connect(host='h', sock=MagicMock(), ssh_config=True)
        hk_inst.load.assert_called_once_with('file_name')
        mock_os.mock_calls == [call('~/.ssh/config'), call('known_hosts_file')]

    @patch('ncclient.transport.ssh.SSHSession.close')
    @patch('paramiko.channel.Channel.recv')
    @patch('selectors.DefaultSelector.select')
    @patch('ncclient.transport.ssh.Session._dispatch_error')
    def test_run_receive_py3(self, mock_error, mock_selector, mock_recv, mock_close):
        mock_selector.return_value = True
        mock_recv.return_value = 0
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._channel = paramiko.Channel("c100")
        obj.run()
        self.assertTrue(
            isinstance(
                mock_error.call_args_list[0][0][0],
                SessionCloseError))

    def test_run_send_py3_10(self):
        self._test_run_send_py3(NetconfBase.BASE_10,
                                lambda msg: msg.encode() + b"]]>]]>")

    def test_run_send_py3_11(self):
        def chunker(msg):
            encmsg = msg.encode()
            chunks = b"\n#%i\n%b\n##\n" % (len(encmsg), encmsg)
            return chunks
        self._test_run_send_py3(NetconfBase.BASE_11, chunker)

    @patch('ncclient.transport.ssh.SSHSession.close')
    @patch('paramiko.channel.Channel.send_ready')
    @patch('paramiko.channel.Channel.send')
    @patch('selectors.DefaultSelector.select')
    @patch('ncclient.transport.ssh.Session._dispatch_error')
    def _test_run_send_py3(self, base, chunker, mock_error, mock_selector,
                           mock_send, mock_ready, mock_close):
        mock_selector.return_value = False
        mock_ready.return_value = True
        mock_send.return_value = -1
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._channel = paramiko.Channel("c100")
        msg = "naïve garçon"
        obj._q.put(msg)
        obj._base = base
        obj.run()
        self.assertEqual(mock_send.call_args_list[0][0][0], chunker(msg))
        self.assertTrue(
            isinstance(
                mock_error.call_args_list[0][0][0],
                SessionCloseError))

    @unittest.skip("test currently non-functional")
    @patch('ncclient.transport.ssh.SSHSession.close')
    @patch('paramiko.channel.Channel.send_ready')
    @patch('paramiko.channel.Channel.send')
    @patch('selectors2.DefaultSelector')
    @patch('ncclient.transport.ssh.Session._dispatch_error')
    def test_run_send_py2(self, mock_error, mock_selector, mock_send, mock_ready, mock_close):
        mock_selector.select.return_value = False
        mock_ready.return_value = True
        mock_send.return_value = -1
        device_handler = JunosDeviceHandler({'name': 'junos'})
        obj = SSHSession(device_handler)
        obj._channel = paramiko.Channel("c100")
        obj._q.put("rpc")
        obj.run()
        self.assertEqual(mock_send.call_args_list[0][0][0], "rpc]]>]]>")
        self.assertTrue(
            isinstance(
                mock_error.call_args_list[0][0][0],
                SessionCloseError))



#! /usr/bin/env python
#
# Retrieve a portion selected by an XPATH expression from the running
# config from the NETCONF server passed on the command line using
# get-config and write the XML configs to files.
#
# $ ./nc03.py broccoli "aaa/authentication/users/user[name='schoenw']"

import sys, os, warnings
warnings.simplefilter("ignore", DeprecationWarning)
from ncclient import manager

def demo(host, user, expr):
    with manager.connect(host=host, port=22, username=user) as m:
        assert(":xpath" in m.server_capabilities)
        c = m.get_config(source='running', filter=('xpath', expr)).data_xml
        with open("%s.xml" % host, 'w') as f:
            f.write(c)

if __name__ == '__main__':
    demo(sys.argv[1], os.getenv("USER"), sys.argv[2])



import os
import unittest
from unittest.mock import patch

import paramiko

from ncclient import manager
from ncclient.transport.ssh import SSHSession
from ncclient.operations.third_party.juniper.rpc import *
from ncclient.operations import RaiseMode
from ncclient.transport.parser import DefaultXMLParser

try:
    import selectors
except ImportError:
    import selectors2 as selectors


class TestSession(unittest.TestCase):

    @patch('ncclient.transport.SSHSession.connected')
    @patch('paramiko.channel.Channel.send_ready')
    @patch('paramiko.channel.Channel.send')
    @patch('ncclient.transport.ssh.SSHSession.close')
    @patch('paramiko.channel.Channel.recv')
    @patch('ncclient.transport.SSHSession')
    @patch('selectors.DefaultSelector.select')
    @patch('ncclient.operations.rpc.uuid4')
    def test_filter_xml_sax_on(self, mock_uuid4, mock_select, mock_session, mock_recv,
                              mock_close, mock_send, mock_send_ready, mock_connected):
        mock_send.return_value = True
        mock_send_ready.return_value = -1
        mock_uuid4.return_value = type('dummy', (), {'urn': "urn:uuid:e0a7abe3-fffa-11e5-b78e-b8e85604f858"})
        device_handler = manager.make_device_handler({'name': 'junos', 'use_filter': True})
        rpc = '<get-software-information/>'
        mock_recv.side_effect = self._read_file('get-software-information.xml')
        session = SSHSession(device_handler)
        session._connected = True
        session._channel = paramiko.Channel("c100")
        session.parser = session._device_handler.get_xml_parser(session)
        obj = ExecuteRpc(session, device_handler, raise_mode=RaiseMode.ALL)
        obj._filter_xml = '<multi-routing-engine-results><multi-routing-engine-item><re-name/></multi-routing-engine-item></multi-routing-engine-results>'
        session.run()
        resp = obj.request(rpc)._NCElement__doc[0]
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/re-name')), 2)
        # as filter_xml is not having software-information, response wont contain it
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/software-information')), 0)

    @patch('ncclient.transport.SSHSession.connected')
    @patch('paramiko.channel.Channel.send_ready')
    @patch('paramiko.channel.Channel.send')
    @patch('ncclient.transport.ssh.SSHSession.close')
    @patch('paramiko.channel.Channel.recv')
    @patch('ncclient.transport.SSHSession')
    @patch('selectors.DefaultSelector.select')
    @patch('ncclient.operations.rpc.uuid4')
    def test_filter_xml_sax_on_junos_rfc_compliant(self, mock_uuid4, mock_select, mock_session, mock_recv,
                              mock_close, mock_send, mock_send_ready, mock_connected):
        mock_send.return_value = True
        mock_send_ready.return_value = -1
        mock_uuid4.return_value = type('dummy', (), {'urn': "urn:uuid:e0a7abe3-fffa-11e5-b78e-b8e85604f858"})
        device_handler = manager.make_device_handler({'name': 'junos', 'use_filter': True})
        rpc = '<get-software-information/>'
        mock_recv.side_effect = self._read_file('get-software-information.xml')
        session = SSHSession(device_handler)
        session._connected = True
        session._channel = paramiko.Channel("c100")
        session.parser = session._device_handler.get_xml_parser(session)
        obj = ExecuteRpc(session, device_handler, raise_mode=RaiseMode.ALL)
        obj._filter_xml = '<multi-routing-engine-results><multi-routing-engine-item><re-name/></multi-routing-engine-item></multi-routing-engine-results>'
        session.run()
        resp = obj.request(rpc)._NCElement__doc[0]
        from lxml import etree
        print(resp)
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/re-name')), 2)
        # as filter_xml is not having software-information, response wont contain it
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/software-information')), 0)
        
    @patch('ncclient.transport.SSHSession.connected')
    @patch('paramiko.channel.Channel.send_ready')
    @patch('paramiko.channel.Channel.send')
    @patch('ncclient.transport.ssh.SSHSession.close')
    @patch('paramiko.channel.Channel.recv')
    @patch('ncclient.transport.SSHSession')
    @patch('selectors.DefaultSelector.select')
    @patch('ncclient.operations.rpc.uuid4')
    def test_filter_xml_delimiter_rpc_reply(self, mock_uuid4, mock_select,
                                            mock_session, mock_recv, mock_close,
                                            mock_send, mock_send_ready,
                                            mock_connected):
        mock_send.return_value = True
        mock_send_ready.return_value = -1
        mock_uuid4.return_value = type('dummy', (), {'urn': "urn:uuid:e0a7abe3-fffa-11e5-b78e-b8e85604f858"})
        device_handler = manager.make_device_handler({'name': 'junos', 'use_filter': True})
        rpc = '<get-software-information/>'
        mock_recv.side_effect = self._read_file('get-software-information.xml')[:-1] + [b"</rpc-reply>]]>", b"]]>"]
        session = SSHSession(device_handler)
        session._connected = True
        session._channel = paramiko.Channel("c100")
        session.parser = session._device_handler.get_xml_parser(session)
        obj = ExecuteRpc(session, device_handler, raise_mode=RaiseMode.ALL)
        obj._filter_xml = '<multi-routing-engine-results><multi-routing-engine-item><re-name/></multi-routing-engine-item></multi-routing-engine-results>'
        session.run()
        resp = obj.request(rpc)._NCElement__doc[0]
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/re-name')), 2)
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/software-information')), 0)

    @patch('ncclient.transport.SSHSession.connected')
    @patch('paramiko.channel.Channel.send_ready')
    @patch('paramiko.channel.Channel.send')
    @patch('ncclient.transport.ssh.SSHSession.close')
    @patch('paramiko.channel.Channel.recv')
    @patch('ncclient.transport.SSHSession')
    @patch('selectors.DefaultSelector.select')
    @patch('ncclient.operations.rpc.uuid4')
    def test_filter_xml_delimiter_multiple_rpc_reply(self, mock_uuid4, mock_select,
                                            mock_session, mock_recv, mock_close,
                                            mock_send, mock_send_ready,
                                            mock_connected):
        mock_send.return_value = True
        mock_send_ready.return_value = -1
        mock_uuid4.return_value = type('dummy', (), {'urn': "urn:uuid:e0a7abe3-fffa-11e5-b78e-b8e85604f858"})
        device_handler = manager.make_device_handler({'name': 'junos', 'use_filter': True})
        rpc = '<get-software-information/>'
        mock_recv.side_effect = self._read_file('get-software-information.xml')[:-1] + [b"</rpc-reply>]]>",
                                                                                        b"]]><rpc-reply>"] + \
                                self._read_file('get-software-information.xml')[1:]
        session = SSHSession(device_handler)
        session._connected = True
        session._channel = paramiko.Channel("c100")
        session.parser = session._device_handler.get_xml_parser(session)
        obj = ExecuteRpc(session, device_handler, raise_mode=RaiseMode.ALL)
        obj._filter_xml = '<multi-routing-engine-results><multi-routing-engine-item><re-name/></multi-routing-engine-item></multi-routing-engine-results>'
        session.run()
        resp = obj.request(rpc)._NCElement__doc[0]
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/re-name')), 2)
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/software-information')), 0)

        @patch('ncclient.transport.SSHSession.connected')
        @patch('paramiko.channel.Channel.send_ready')
        @patch('paramiko.channel.Channel.send')
        @patch('ncclient.transport.ssh.SSHSession.close')
        @patch('paramiko.channel.Channel.recv')
        @patch('ncclient.transport.SSHSession')
        @patch('selectors.DefaultSelector.select')
        @patch('ncclient.operations.rpc.uuid4')
        def test_filter_xml_delimiter_multiple_rpc_in_parallel(self, mock_uuid4, mock_select,
                                                               mock_session, mock_recv, mock_close,
                                                               mock_send, mock_send_ready,
                                                               mock_connected):
            mock_send.return_value = True
            mock_send_ready.return_value = -1
            mock_uuid4.side_effect = [type('xyz', (), {'urn': "urn:uuid:ddef40cb-5745-481d-974d-7188f9f2bb33"}),
                                      type('pqr', (), {'urn': "urn:uuid:549ef9d1-024a-4fd0-88bf-047d25f0870d"})]
            device_handler = manager.make_device_handler({'name': 'junos', 'use_filter': True})
            rpc = '<get-software-information/>'
            mock_recv.side_effect = [b"""
            <rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" xmlns:junos="http://xml.juniper.net/junos/19.2I0/junos" xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0" message-id="urn:uuid:ddef40cb-5745-481d-974d-7188f9f2bb33">
    <ospf-neighbor-information xmlns="http://xml.juniper.net/junos/19.2I0/junos-routing">
    <ospf-neighbor>
    <neighbor-address>13.1.1.2</neighbor-address>
    <interface-name>ge-0/0/0.1</interface-name>
    <ospf-neighbor-state>Exchange</ospf-neighbor-state>
    <neighbor-id>2.2.2.2</neighbor-id>
    <neighbor-priority>128</neighbor-priority>
    <activity-timer>36</activity-timer>
    <ospf-area>0.0.0.0</ospf-area>
    <options>0x52</options>
    <dr-address>13.1.1.1</dr-address>
    <bdr-address>13.1.1.2</bdr-address>
    <neighbor-up-time junos:seconds="17812">
    04:56:52
    </neighbor-up-time>
    <neighbor-adjacency-time junos:seconds="17812">
    04:56:52
    </neighbor-adjacency-time>
    <master-slave>slave</master-slave>
    <sequence-number>0x204b6fd</sequence-number>
    <dbd-retransmit-time>3</dbd-retransmit-time>
    <lsreq-retransmit-time>0</lsreq-retransmit-time>
    <lsa-list>
      Link state retransmission list:
        Type      LSA ID           Adv rtr          Seq
       Router    1.1.1.1          1.1.1.1          0x80000019
       OpaqArea  1.0.0.1          1.1.1.1          0x80000011
       Router    3.3.3.3          3.3.3.3          0x80000004
       Network   23.1.1.2         3.3.3.3          0x80000001
       OpaqArea  1.0.0.1          2.2.2.2          0x80000002
       OpaqArea  1.0.0.1          3.3.3.3          0x80000002
       OpaqArea  1.0.0.3          1.1.1.1          0x80000002
       OpaqArea  1.0.0.3          3.3.3.3          0x80000001
       OpaqArea  1.0.0.4          2.2.2.2          0x80000001
    </lsa-list>
    <ospf-neighbor-topology>
    <ospf-topology-name>default</ospf-topology-name>
    <ospf-topology-id>0</ospf-topology-id>
    <ospf-neighbor-topology-state>Forward Only</ospf-neighbor-topology-state>
    </ospf-neighbor-topology>
    </ospf-neighbor>
    </ospf-neighbor-information>
    </rpc-reply>]]>]]><rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" xmlns:junos="http://xml.juniper.net/junos/19.2I0/junos" xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0" message-id="urn:uuid:549ef9d1-024a-4fd0-88bf-047d25f0870d">
    <pfe-statistics>
    <pfe-traffic-statistics>
    <pfe-input-packets>22450</pfe-input-packets>
    <input-pps>0</input-pps>
    <pfe-output-packets>31992</pfe-output-packets>
    <output-pps>0</output-pps>
    <pfe-fabric-input>0</pfe-fabric-input>
    <pfe-fabric-input-pps>0</pfe-fabric-input-pps>
    <pfe-fabric-output>0</pfe-fabric-output>
    <pfe-fabric-output-pps>0</pfe-fabric-output-pps>
    </pfe-traffic-statistics></pfe-statistics></rpc-reply>]]>]]>"""]
            session = SSHSession(device_handler)
            session._connected = True
            session._channel = paramiko.Channel("c100")
            session.parser = session._device_handler.get_xml_parser(session)
            obj = ExecuteRpc(session, device_handler, raise_mode=RaiseMode.ALL)
            obj = ExecuteRpc(session, device_handler, raise_mode=RaiseMode.ALL)
            obj._filter_xml = '<multi-routing-engine-results><multi-routing-engine-item><re-name/></multi-routing-engine-item></multi-routing-engine-results>'
            session.run()
            resp = obj.request(rpc)._NCElement__doc[0]
            self.assertEqual(len(resp.xpath('pfe-traffic-statistics')), 1)

    @patch('ncclient.transport.SSHSession.connected')
    @patch('paramiko.channel.Channel.send_ready')
    @patch('paramiko.channel.Channel.send')
    @patch('ncclient.transport.ssh.SSHSession.close')
    @patch('paramiko.channel.Channel.recv')
    @patch('ncclient.transport.SSHSession')
    @patch('selectors.DefaultSelector.select')
    @patch('ncclient.operations.rpc.uuid4')
    def test_filter_xml_delimiter_splited_rpc_reply(self, mock_uuid4, mock_select,
                                            mock_session, mock_recv, mock_close,
                                            mock_send, mock_send_ready,
                                            mock_connected):
        mock_send.return_value = True
        mock_send_ready.return_value = -1
        mock_uuid4.return_value = type('dummy', (), {'urn': "urn:uuid:e0a7abe3-fffa-11e5-b78e-b8e85604f858"})
        device_handler = manager.make_device_handler({'name': 'junos', 'use_filter': True})
        rpc = '<get-software-information/>'
        mock_recv.side_effect = self._read_file('get-software-information.xml')[:-1] + [b"</rpc", b"-reply>]]>",
                                                                                        b"]]><rpc-reply>"] + \
                                self._read_file('get-software-information.xml')[1:]
        session = SSHSession(device_handler)
        session._connected = True
        session._channel = paramiko.Channel("c100")
        session.parser = session._device_handler.get_xml_parser(session)
        obj = ExecuteRpc(session, device_handler, raise_mode=RaiseMode.ALL)
        obj._filter_xml = '<multi-routing-engine-results><multi-routing-engine-item><re-name/></multi-routing-engine-item></multi-routing-engine-results>'
        session.run()
        resp = obj.request(rpc)._NCElement__doc[0]
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/re-name')), 2)
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/software-information')), 0)

    @patch('ncclient.transport.SSHSession.connected')
    @patch('paramiko.channel.Channel.send_ready')
    @patch('paramiko.channel.Channel.send')
    @patch('ncclient.transport.ssh.SSHSession.close')
    @patch('paramiko.channel.Channel.recv')
    @patch('ncclient.transport.SSHSession')
    @patch('selectors.DefaultSelector.select')
    @patch('ncclient.operations.rpc.uuid4')
    def test_use_filter_xml_without_sax_input(self, mock_uuid4, mock_select,
                                              mock_session, mock_recv,
                                              mock_close, mock_send,
                                              mock_send_ready,
                                              mock_connected):
        mock_send.return_value = True
        mock_send_ready.return_value = -1
        mock_uuid4.return_value = type('dummy', (), {'urn': "urn:uuid:e0a7abe3-fffa-11e5-b78e-b8e85604f858"})
        device_handler = manager.make_device_handler({'name': 'junos', 'use_filter': True})
        rpc = '<get-software-information/>'
        mock_recv.side_effect = self._read_file('get-software-information.xml')
        session = SSHSession(device_handler)
        session._connected = True
        session._channel = paramiko.Channel("c100")
        session.parser = session._device_handler.get_xml_parser(session)
        obj = ExecuteRpc(session, device_handler, raise_mode=RaiseMode.ALL)
        obj._filter_xml = None
        session.run()
        resp = obj.request(rpc)._NCElement__doc[0]
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/re-name')), 2)
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/software-information')), 2)

    @patch('ncclient.transport.SSHSession.connected')
    @patch('paramiko.channel.Channel.send_ready')
    @patch('paramiko.channel.Channel.send')
    @patch('ncclient.transport.ssh.SSHSession.close')
    @patch('paramiko.channel.Channel.recv')
    @patch('ncclient.transport.SSHSession')
    @patch('selectors.DefaultSelector.select')
    @patch('ncclient.operations.rpc.uuid4')
    def test_use_filter_False(self, mock_uuid4, mock_select,
                                              mock_session, mock_recv,
                                              mock_close, mock_send,
                                              mock_send_ready,
                                              mock_connected):
        mock_send.return_value = True
        mock_send_ready.return_value = -1
        mock_uuid4.return_value = type('dummy', (), {'urn': "urn:uuid:e0a7abe3-fffa-11e5-b78e-b8e85604f858"})
        device_handler = manager.make_device_handler({'name': 'junos', 'use_filter': False})
        rpc = '<get-software-information/>'
        mock_recv.side_effect = self._read_file('get-software-information.xml')
        session = SSHSession(device_handler)
        session._connected = True
        session._channel = paramiko.Channel("c100")
        session.parser = session._device_handler.get_xml_parser(session)
        obj = ExecuteRpc(session, device_handler, raise_mode=RaiseMode.ALL)
        obj._filter_xml = '<multi-routing-engine-results><multi-routing-engine-item><re-name/></multi-routing-engine-item></multi-routing-engine-results>'
        session.run()
        resp = obj.request(rpc)._NCElement__doc[0]
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/re-name')), 2)
        # as filter_xml is not having software-information, response wont contain it
        self.assertEqual(len(resp.xpath('multi-routing-engine-item/software-information')), 2)
        self.assertIsInstance(session.parser, DefaultXMLParser)

    def _read_file(self, fname):
        fpath = os.path.join(os.path.dirname(__file__),
                             'rpc-reply', fname)
        lines = []
        with open(fpath, "rb") as fp:
            lines = fp.readlines()
        return lines



import unittest
from xml.etree import ElementTree

try:
    from unittest.mock import patch  # Python 3.4 and later
except ImportError:
    from mock import patch
from ncclient import manager
import ncclient.transport
from ncclient.operations.third_party.sros.rpc import *
from ncclient.operations import RaiseMode


class TestRPC(unittest.TestCase):
    def setUp(self):
        self.device_handler = manager.make_device_handler({'name': 'sros'})

    @patch('ncclient.operations.third_party.sros.rpc.RPC._request')
    def test_MdCliRawCommand(self, mock_request):
        mock_request.return_value = 'sros'
        expected = 'sros'
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = MdCliRawCommand(session, self.device_handler)
        command = 'show version'
        actual = obj.request(command=command)
        self.assertEqual(expected, actual)

    @patch('ncclient.transport.SSHSession')
    @patch('ncclient.operations.third_party.sros.rpc.RPC._request')
    @patch('ncclient.operations.third_party.sros.rpc.RPC._assert')
    def test_commit(self, mock_assert, mock_request, mock_session):
        device_handler = manager.make_device_handler({'name': 'sros'})
        session = ncclient.transport.SSHSession(device_handler)
        obj = Commit(session, device_handler, raise_mode=RaiseMode.ALL)

        obj.request(confirmed=True, comment="This is a comment", timeout="50")

        node = new_ele("commit")
        sub_ele(node, "comment",
                attrs={'xmlns': "urn:nokia.com:sros:ns:yang:sr:ietf-netconf-augments"}).text = "This is a comment"
        sub_ele(node, "confirmed")
        sub_ele(node, "confirm-timeout").text = "50"

        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)

        self.assertEqual(call, xml)



# Copyright 2009 Shikhar Bhushan
# Copyright 2011 Leonidas Poulopoulos
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module is a thin layer of abstraction around the library.
It exposes all core functionality.
"""

from ncclient import operations
from ncclient import transport
import socket
import logging
import functools

from ncclient.xml_ import *

logger = logging.getLogger('ncclient.manager')

OPERATIONS = {
    "get": operations.Get,
    "get_config": operations.GetConfig,
    "get_schema": operations.GetSchema,
    "dispatch": operations.Dispatch,
    "edit_config": operations.EditConfig,
    "copy_config": operations.CopyConfig,
    "validate": operations.Validate,
    "commit": operations.Commit,
    "discard_changes": operations.DiscardChanges,
    "cancel_commit": operations.CancelCommit,
    "delete_config": operations.DeleteConfig,
    "lock": operations.Lock,
    "unlock": operations.Unlock,
    "create_subscription": operations.CreateSubscription,
    "close_session": operations.CloseSession,
    "kill_session": operations.KillSession,
    "poweroff_machine": operations.PoweroffMachine,
    "reboot_machine": operations.RebootMachine,
    "rpc": operations.GenericRPC,
}

"""
Dictionary of base method names and corresponding :class:`~ncclient.operations.RPC`
subclasses. It is used to lookup operations, e.g. `get_config` is mapped to
:class:`~ncclient.operations.GetConfig`. It is thus possible to add additional
operations to the :class:`Manager` API.
"""


def make_device_handler(device_params, ignore_errors=None):
    """
    Create a device handler object that provides device specific parameters and
    functions, which are called in various places throughout our code.

    If no device_params are defined or the "name" in the parameter dict is not
    known then a default handler will be returned.

    """
    if device_params is None:
        device_params = {}

    handler = device_params.get('handler', None)
    if handler:
        return handler(device_params, ignore_errors)

    device_name = device_params.get("name", "default")
    # Attempt to import device handler class. All device handlers are
    # in a module called "ncclient.devices.<devicename>" and in a class named
    # "<devicename>DeviceHandler", with the first letter capitalized.
    class_name          = "%sDeviceHandler" % device_name.capitalize()
    devices_module_name = "ncclient.devices.%s" % device_name
    dev_module_obj      = __import__(devices_module_name)
    handler_module_obj  = getattr(getattr(dev_module_obj, "devices"), device_name)
    class_obj           = getattr(handler_module_obj, class_name)
    handler_obj         = class_obj(device_params, ignore_errors)
    return handler_obj


def _extract_device_params(kwds):
    device_params = kwds.pop("device_params", None)

    return device_params

def _extract_errors_params(kwds):
    errors_params = kwds.pop("errors_params", {})
    return (
        errors_params.get("ignore_errors", []),
        errors_params.get("raise_mode", operations.RaiseMode.ALL)
    )

def _extract_manager_params(kwds):
    manager_params = kwds.pop("manager_params", {})

    # To maintain backward compatibility
    if 'timeout' not in manager_params and 'timeout' in kwds:
        manager_params['timeout'] = kwds['timeout']
    return manager_params

def _extract_nc_params(kwds):
    nc_params = kwds.pop("nc_params", {})

    return nc_params

def connect_ssh(*args, **kwds):
    """Initialize a :class:`Manager` over the SSH transport.
    For documentation of arguments see :meth:`ncclient.transport.SSHSession.connect`.

    The underlying :class:`ncclient.transport.SSHSession` is created with
    :data:`CAPABILITIES`. All the provided arguments are passed directly to its
    implementation of :meth:`~ncclient.transport.SSHSession.connect`.

    To customize the :class:`Manager`, add a `manager_params` dictionary in connection
    parameters (e.g. `manager_params={'timeout': 60}` for a bigger RPC timeout parameter)

    To invoke advanced vendor related operation add
    `device_params={'name': '<vendor_alias>'}` in connection parameters. For the time,
    'junos' and 'nexus' are supported for Juniper and Cisco Nexus respectively.

    A custom device handler can be provided with
    `device_params={'handler':<handler class>}` in connection parameters.

    To specify errors to be be ignored in the RPC reply add
    `errors_params={'ignore_errors': ['error pattern to ignore']}` in connection parameters.

    To control the error raising mode add `raise_mode` in the `errors_params`
    (e.g. `errors_params={'raise_mode': 0}` for ignoring all RPC errors)
    See :class:`ncclient.operations.rpc.RaiseMode` for valid values of `raise_mode`

    """
    # Extract device/manager/netconf parameter dictionaries, if they were passed into this function.
    # Remove them from kwds (which should keep only session.connect() parameters).
    device_params = _extract_device_params(kwds)
    manager_params = _extract_manager_params(kwds)
    nc_params = _extract_nc_params(kwds)
    ignore_errors, raise_mode = _extract_errors_params(kwds)
    manager_params["raise_mode"] = raise_mode

    device_handler = make_device_handler(device_params, ignore_errors)
    device_handler.add_additional_ssh_connect_params(kwds)
    device_handler.add_additional_netconf_params(nc_params)
    session = transport.SSHSession(device_handler)

    try:
       session.connect(*args, **kwds)
    except Exception as ex:
        if session.transport:
            session.close()
        raise
    return Manager(session, device_handler, **manager_params)

def connect_libssh(*args, **kwargs):
    """Initialize a :class:`Manager` over the LibSSH transport."""
    if not hasattr(transport, 'libssh'):
        raise ValueError("LibSSH transport is not available, install 'ssh-python' package.")

    device_params = _extract_device_params(kwargs)
    manager_params = _extract_manager_params(kwargs)
    nc_params = _extract_nc_params(kwargs)
    ignore_errors, raise_mode = _extract_errors_params(kwargs)
    manager_params["raise_mode"] = raise_mode

    device_handler = make_device_handler(device_params, ignore_errors)
    device_handler.add_additional_ssh_connect_params(kwargs)
    device_handler.add_additional_netconf_params(nc_params)
    session = transport.libssh.LibSSHSession(device_handler)

    session.connect(*args, **kwargs)
    return Manager(session, device_handler, **manager_params)

def connect_tls(*args, **kwargs):
    """Initialize a :class:`Manager` over the TLS transport."""
    device_params = _extract_device_params(kwargs)
    manager_params = _extract_manager_params(kwargs)
    nc_params = _extract_nc_params(kwargs)
    ignore_errors, raise_mode = _extract_errors_params(kwargs)
    manager_params["raise_mode"] = raise_mode

    device_handler = make_device_handler(device_params, ignore_errors)
    device_handler.add_additional_netconf_params(nc_params)
    session = transport.TLSSession(device_handler)

    session.connect(*args, **kwargs)

    return Manager(session, device_handler, **manager_params)

def connect_uds(*args, **kwargs):
    """Initialize a :class:`Manager` over the Unix Socket transport."""
    device_params = _extract_device_params(kwargs)
    manager_params = _extract_manager_params(kwargs)
    nc_params = _extract_nc_params(kwargs)
    ignore_errors, raise_mode = _extract_errors_params(kwargs)
    manager_params["raise_mode"] = raise_mode

    device_handler = make_device_handler(device_params, ignore_errors)
    device_handler.add_additional_netconf_params(nc_params)
    session = transport.UnixSocketSession(device_handler)

    session.connect(*args, **kwargs)

    return Manager(session, device_handler, **manager_params)

def connect_ioproc(*args, **kwds):
    device_params = _extract_device_params(kwds)
    manager_params = _extract_manager_params(kwds)
    ignore_errors, raise_mode = _extract_errors_params(kwds)
    manager_params["raise_mode"] = raise_mode

    if device_params:
        import_string = 'ncclient.transport.third_party.'
        import_string += device_params['name'] + '.ioproc'
        third_party_import = __import__(import_string, fromlist=['IOProc'])

    device_handler = make_device_handler(device_params, ignore_errors)

    session = third_party_import.IOProc(device_handler)
    session.connect()

    return Manager(session, device_handler, **manager_params)


def connect(*args, **kwds):
    if "host" in kwds:
        host = kwds["host"]
        device_params = kwds.get('device_params', {})
        if host == 'localhost' and device_params.get('name') == 'junos' \
                and device_params.get('local'):
            return connect_ioproc(*args, **kwds)
        elif kwds.pop("use_libssh", False):
            return connect_libssh(*args, **kwds)
        else:
            return connect_ssh(*args, **kwds)

def call_home(*args, **kwds):
    host = kwds["host"]
    port = kwds.get("port",4334)
    srv_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    srv_socket.bind((host, port))
    srv_socket.settimeout(kwds.get("timeout", 10))
    srv_socket.listen()

    sock, remote_host = srv_socket.accept()
    logger.info('Callhome connection initiated from remote host {0}'.format(remote_host))
    kwds['sock'] = sock
    return connect_ssh(*args, **kwds)

class Manager:

    """
    For details on the expected behavior of the operations and their
        parameters refer to :rfc:`6241`.

    Manager instances are also context managers so you can use it like this::

        with manager.connect("host") as m:
            # do your stuff

    ... or like this::

        m = manager.connect("host")
        try:
            # do your stuff
        finally:
            m.close_session()
    """

   # __metaclass__ = OpExecutor


    HUGE_TREE_DEFAULT = False
    """Default for `huge_tree` support for XML parsing of RPC replies (defaults to False)"""

    def __init__(self, session, device_handler, timeout=30, raise_mode=operations.RaiseMode.ALL):
        self._session = session
        self._async_mode = False
        self._timeout = timeout
        self._raise_mode = raise_mode
        self._huge_tree = self.HUGE_TREE_DEFAULT
        self._device_handler = device_handler
        self._vendor_operations = {}
        if device_handler:
            self._vendor_operations.update(device_handler.add_additional_operations())

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close_session()
        return False

    def __set_timeout(self, timeout):
        self._timeout = timeout

    def __set_async_mode(self, mode):
        self._async_mode = mode

    def __set_raise_mode(self, mode):
        assert(mode in (operations.RaiseMode.NONE, operations.RaiseMode.ERRORS, operations.RaiseMode.ALL))
        self._raise_mode = mode

    def execute(self, cls, *args, **kwds):
        return cls(self._session,
                   device_handler=self._device_handler,
                   async_mode=self._async_mode,
                   timeout=self._timeout,
                   raise_mode=self._raise_mode,
                   huge_tree=self._huge_tree).request(*args, **kwds)

    def locked(self, target):
        """Returns a context manager for a lock on a datastore, where
        *target* is the name of the configuration datastore to lock, e.g.::

            with m.locked("running"):
                # do your stuff

        ... instead of::

            m.lock("running")
            try:
                # do your stuff
            finally:
                m.unlock("running")
        """
        return operations.LockContext(self._session, self._device_handler, target)

    def scp(self):
        return self._session.scp()

    def session(self):
        raise NotImplementedError

    def __getattr__(self, method):
        if method in self._vendor_operations:
            return functools.partial(self.execute, self._vendor_operations[method])
        elif method in OPERATIONS:
            return functools.partial(self.execute, OPERATIONS[method])
        else:
            """Parse args/kwargs correctly in order to build XML element"""
            def _missing(*args, **kwargs):
                m = method.replace('_', '-')
                root = new_ele(m)
                if args:
                    for arg in args:
                        sub_ele(root, arg)
                r = self.rpc(root)
                return r
            return _missing

    def take_notification(self, block=True, timeout=None):
        """Attempt to retrieve one notification from the queue of received
        notifications.

        If block is True, the call will wait until a notification is
        received.

        If timeout is a number greater than 0, the call will wait that
        many seconds to receive a notification before timing out.

        If there is no notification available when block is False or
        when the timeout has elapse, None will be returned.

        Otherwise a :class:`~ncclient.operations.notify.Notification`
        object will be returned.
        """
        return self._session.take_notification(block, timeout)

    @property
    def client_capabilities(self):
        """:class:`~ncclient.capabilities.Capabilities` object representing
        the client's capabilities."""
        return self._session._client_capabilities

    @property
    def server_capabilities(self):
        """:class:`~ncclient.capabilities.Capabilities` object representing
        the server's capabilities."""
        return self._session._server_capabilities

    @property
    def channel_id(self):
        return self._session._channel_id

    @property
    def channel_name(self):
        return self._session._channel_name

    @property
    def session_id(self):
        """`session-id` assigned by the NETCONF server."""
        return self._session.id

    @property
    def connected(self):
        """Whether currently connected to the NETCONF server."""
        return self._session.connected

    async_mode = property(fget=lambda self: self._async_mode,
                          fset=__set_async_mode)
    """Specify whether operations are executed asynchronously (`True`) or
    synchronously (`False`) (the default)."""

    timeout = property(fget=lambda self: self._timeout, fset=__set_timeout)
    """Specify the timeout for synchronous RPC requests."""

    raise_mode = property(fget=lambda self: self._raise_mode,
                          fset=__set_raise_mode)
    """Specify which errors are raised as :exc:`~ncclient.operations.RPCError`
    exceptions. Valid values are the constants defined in
    :class:`~ncclient.operations.RaiseMode`.
    The default value is :attr:`~ncclient.operations.RaiseMode.ALL`."""

    @property
    def huge_tree(self):
        """Whether `huge_tree` support for XML parsing of RPC replies is enabled (default=False)
        The default value is configurable through :attr:`~ncclient.manager.Manager.HUGE_TREE_DEFAULT`"""
        return self._huge_tree

    @huge_tree.setter
    def huge_tree(self, x):
        self._huge_tree = x



# -*- coding: utf-8 -*-
# by yangxufeng.zhao

from ncclient.xml_ import *

from ncclient.operations.rpc import RPC


class CLI(RPC):
    def request(self, command=None):
        """command text
        view: Execution user view exec
              Configuration system view exec
        """
        # node = new_ele("execute-cli")
        node = new_ele("execute-cli", attrs={"xmlns":HW_PRIVATE_NS})
        node.append(validated_element(command))
        return self._request(node)


class Action(RPC):
    "`execute-action` RPC"
    def request(self, action=None):
        node = new_ele("execute-action", attrs={"xmlns":HW_PRIVATE_NS})
        node.append(validated_element(action))
        return self._request(node)



import unittest
from ncclient.devices.huawei import *


capabilities = ['urn:ietf:params:netconf:base:1.0',
                'urn:ietf:params:netconf:base:1.1',
                'urn:ietf:params:netconf:capability:writable-running:1.0',
                'urn:ietf:params:netconf:capability:candidate:1.0',
                'urn:ietf:params:netconf:capability:confirmed-commit:1.0',
                'urn:ietf:params:netconf:capability:rollback-on-error:1.0',
                'urn:ietf:params:netconf:capability:startup:1.0',
                'urn:ietf:params:netconf:capability:url:1.0?scheme=http,ftp,file,https,sftp',
                'urn:ietf:params:netconf:capability:validate:1.0',
                'urn:ietf:params:netconf:capability:xpath:1.0',
                'urn:ietf:params:netconf:capability:notification:1.0',
                'urn:ietf:params:netconf:capability:interleave:1.0',
                'urn:ietf:params:netconf:capability:with-defaults:1.0',
                'http://www.huawei.com/netconf/capability/execute-cli/1.0',
                'http://www.huawei.com/netconf/capability/action/1.0',
                'http://www.huawei.com/netconf/capability/active/1.0',
                'http://www.huawei.com/netconf/capability/discard-commit/1.0',
                'http://www.huawei.com/netconf/capability/exchange/1.0']


class TestHuaweiDevice(unittest.TestCase):

    def setUp(self):
        self.obj = HuaweiDeviceHandler({'name': 'huawei'})

    def test_add_additional_operations(self):
        expected = dict()
        expected['cli'] = CLI
        expected['action'] = Action
        self.assertDictEqual(expected, self.obj.add_additional_operations())

    def test_handle_raw_dispatch(self):
        expected = 'hello'
        self.assertEqual(expected, self.obj.handle_raw_dispatch('hello'))

    def test_get_capabilities(self):
        self.assertListEqual(capabilities, self.obj.get_capabilities())

    def test_get_xml_base_namespace_dict(self):
        expected = {None: BASE_NS_1_0}
        self.assertDictEqual(expected, self.obj.get_xml_base_namespace_dict())

    def test_get_xml_extra_prefix_kwargs(self):
        expected = dict()
        expected['nsmap'] = self.obj.get_xml_base_namespace_dict()
        self.assertDictEqual(expected, self.obj.get_xml_extra_prefix_kwargs())

    def test_perform_qualify_check(self):
        self.assertFalse(self.obj.perform_qualify_check())



from ncclient.operations.edit import *
import unittest
try:
    from unittest.mock import patch, MagicMock  # Python 3.4 and later
except ImportError:
    from mock import patch, MagicMock
from ncclient import manager
import ncclient.manager
import ncclient.transport
from ncclient.xml_ import *
from ncclient.operations import RaiseMode
from xml.etree import ElementTree
from ncclient.operations.errors import MissingCapabilityError
import copy


class TestEdit(unittest.TestCase):

    def setUp(self):
        self.device_handler = manager.make_device_handler({'name': 'junos'})

    @patch('ncclient.operations.edit.RPC._request')
    def test_edit_config(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = EditConfig(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        root = new_ele('config')
        configuration = sub_ele(root, 'configuration')
        system = sub_ele(configuration, 'system')
        location = sub_ele(system, 'location')
        sub_ele(location, 'building').text = "Main Campus, A"
        sub_ele(location, 'floor').text = "5"
        sub_ele(location, 'rack').text = "27"
        obj.request(copy.deepcopy(root))
        node = new_ele("edit-config")
        node.append(util.datastore_or_url("target", "candidate"))
        node.append(validated_element(root, ("config", qualify("config"))))
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.edit.RPC._request')
    def test_edit_config_2(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = [":rollback-on-error", ":validate"]
        obj = EditConfig(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        root = """
            system {
                host-name foo-bar;
            }
            """
        obj.request(copy.deepcopy(root),
                    format="text",
                    target="running",
                    error_option="rollback-on-error",
                    default_operation="merge",
                    test_option="test-then-set")
        node = new_ele("edit-config")
        node.append(util.datastore_or_url("target", "running"))
        sub_ele(node, "default-operation").text = "merge"
        sub_ele(node, "test-option").text = "test-then-set"
        sub_ele(node, "error-option").text = "rollback-on-error"
        config_text = sub_ele(node, "config-text")
        sub_ele(config_text, "configuration-text").text = root
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.edit.RPC._request')
    @patch('ncclient.operations.edit.RPC._assert')
    def test_edit_config_valid_url(self, mock_assert, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = EditConfig(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        config = "https://www.xxx.org/index.html"
        obj.request(config, format='url')
        node = new_ele("edit-config")
        node.append(util.datastore_or_url("target", "candidate"))
        sub_ele(node, "url").text = config

        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.edit.RPC._request')
    @patch('ncclient.operations.edit.RPC._assert')
    def test_edit_config_invalid_url(self, mock_assert, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = EditConfig(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        config = "Invalid URL"
        self.assertRaises(OperationError, obj.request,
                          config, format='url')

    def test_edit_config_invalid_arguments_exception(self):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = [":rollback-on-error", ":validate", ":validate:1.1"]
        obj = EditConfig(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        root = """
            system {
                host-name foo-bar;
            }
            """
        #invalid argument "default_operation"
        self.assertRaises(OperationError, obj.request,
                          copy.deepcopy(root),
                          format="xml",
                          target="running",
                          error_option="stop-on-error",
                          default_operation="create",
                          test_option="test-then-set")
        #invalid argument "error_option"
        self.assertRaises(OperationError, obj.request,
                          copy.deepcopy(root),
                          format="xml",
                          target="running",
                          error_option="commit-on-error",
                          default_operation="merge",
                          test_option="test-then-set")
        #invalid argument "test_option"
        self.assertRaises(OperationError, obj.request,
                          copy.deepcopy(root),
                          format="xml",
                          target="running",
                          error_option="stop-on-error",
                          default_operation="merge",
                          test_option="test")

    def test_edit_config_validate_capability_exception(self):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = [":validate"]
        obj = EditConfig(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        root = """
            system {
                host-name foo-bar;
            }
            """
        self.assertRaises(MissingCapabilityError, obj.request,
                          copy.deepcopy(root),
                          format="xml",
                          target="running",
                          error_option="stop-on-error",
                          default_operation="merge",
                          test_option="test-only")

    @patch('ncclient.operations.RPC._request')
    def test_delete_config(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = DeleteConfig(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        obj.request("candidate")
        node = new_ele("delete-config")
        node.append(util.datastore_or_url("target", "candidate"))
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.RPC._request')
    def test_copy_config(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = CopyConfig(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        obj.request("candidate", "running")
        node = new_ele("copy-config")
        node.append(util.datastore_or_url("target", "running"))
        node.append(util.datastore_or_url("source", "candidate"))
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.RPC._request')
    def test_copy_config_2(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        obj = CopyConfig(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        source = new_ele('source')
        config = sub_ele(source, 'config')
        configuration = sub_ele(config, 'configuration')
        system = sub_ele(configuration, 'system')
        location = sub_ele(system, 'location')
        sub_ele(location, 'building').text = "Main Campus, A"
        sub_ele(location, 'floor').text = "5"
        sub_ele(location, 'rack').text = "27"
        obj.request(copy.deepcopy(source), "candidate")
        node = new_ele("copy-config")
        node.append(util.datastore_or_url("target", "candidate"))
        node.append(validated_element(source, ("source", qualify("source"))))
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.RPC._request')
    def test_validate_config(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = [':validate']
        obj = Validate(session, self.device_handler, raise_mode=RaiseMode.ALL)
        cfg_data = new_ele("config")
        sub_ele(cfg_data, "data")
        obj.request(cfg_data)
        node = new_ele("validate")
        src = sub_ele(node, "source")
        cfg = sub_ele(src, "config")
        sub_ele(cfg, "data")
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.RPC._request')
    def test_validate_datastore(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = [':validate']
        obj = Validate(session, self.device_handler, raise_mode=RaiseMode.ALL)
        obj.request("candidate")
        node = new_ele("validate")
        src = sub_ele(node, "source")
        sub_ele(src, "candidate")
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.RPC._request')
    def test_commit(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = [':candidate', ":confirmed-commit"]
        obj = Commit(session, self.device_handler, raise_mode=RaiseMode.ALL)
        obj.request(confirmed=True, timeout="0")
        node = new_ele("commit")
        sub_ele(node, "confirmed")
        sub_ele(node, "confirm-timeout").text = "0"
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.RPC._request')
    def test_commit_exception(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = [':candidate']
        obj = Commit(session, self.device_handler, raise_mode=RaiseMode.ALL)
        self.assertRaises(MissingCapabilityError,
            obj.request, confirmed=True, timeout="0")

    @patch('ncclient.operations.RPC._request')
    def test_discard_changes(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = [':candidate', ":confirmed-commit"]
        obj = DiscardChanges(
            session,
            self.device_handler,
            raise_mode=RaiseMode.ALL)
        obj.request()
        node = new_ele("discard-changes")
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)

    @patch('ncclient.operations.RPC._request')
    def test_cancel_commit(self, mock_request):
        session = ncclient.transport.SSHSession(self.device_handler)
        session._server_capabilities = [':candidate', ":confirmed-commit"]
        obj = CancelCommit(session, self.device_handler, raise_mode=RaiseMode.ALL)
        obj.request(persist_id="foo")
        node = new_ele("cancel-commit")
        sub_ele(node, "persist-id").text = "foo"
        xml = ElementTree.tostring(node)
        call = mock_request.call_args_list[0][0][0]
        call = ElementTree.tostring(call)
        self.assertEqual(call, xml)



# Copyright 2018 Nitin Kumar

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys
import re

try:
    import selectors
except ImportError:
    import selectors2 as selectors

from io import BytesIO as StringIO
from xml.sax.handler import ContentHandler

from ncclient.transport.errors import NetconfFramingError
from ncclient.transport.session import NetconfBase
from ncclient.logging_ import SessionLoggerAdapter
from ncclient.operations.errors import OperationError
from ncclient.transport import SessionListener

import logging
logger = logging.getLogger("ncclient.transport.parser")


PORT_NETCONF_DEFAULT = 830
PORT_SSH_DEFAULT = 22

BUF_SIZE = 4096
# v1.0: RFC 4742
MSG_DELIM = "]]>]]>"
MSG_DELIM_LEN = len(MSG_DELIM)
# v1.1: RFC 6242
END_DELIM = '\n##\n'

TICK = 0.1

#
# Define delimiters for chunks and messages for netconf 1.1 chunk enoding.
# When matched:
#
# * result.group(0) will contain whole matched string
# * result.group(1) will contain the digit string for a chunk
# * result.group(2) will be defined if '##' found
#
RE_NC11_DELIM = re.compile(r'\n(?:#([0-9]+)|(##))\n')

def textify(buf):
    return buf.decode('UTF-8')


class SAXParserHandler(SessionListener):

    def __init__(self, session):
        self._session = session

    def callback(self, root, raw):
        if type(self._session.parser) == DefaultXMLParser:
            self._session.parser = self._session._device_handler.get_xml_parser(self._session)

    def errback(self, _):
        pass


class SAXFilterXMLNotFoundError(OperationError):
    def __init__(self, rpc_listener):
        self._listener = rpc_listener

    def __str__(self):
        return "SAX filter input xml not provided for listener: %s" % self._listener


class DefaultXMLParser:

    def __init__(self, session):
        """
        DOM Parser

        :param session: ssh session object
        """
        self._session = session
        self._parsing_pos10 = 0
        self.logger = SessionLoggerAdapter(logger, {'session': self._session})

    def parse(self, data):
        """
        parse incoming RPC response from networking device.

        :param data: incoming RPC data from device
        :return: None
        """
        if data:
            self._session._buffer.seek(0, os.SEEK_END)
            self._session._buffer.write(data)
            if self._session._base == NetconfBase.BASE_11:
                self._parse11()
            else:
                self._parse10()

    def _parse10(self):

        """Messages are delimited by MSG_DELIM. The buffer could have grown by
        a maximum of BUF_SIZE bytes everytime this method is called. Retains
        state across method calls and if a chunk has been read it will not be
        considered again."""

        self.logger.debug("parsing netconf v1.0")
        buf = self._session._buffer
        buf.seek(self._parsing_pos10)
        if MSG_DELIM in buf.read().decode('UTF-8'):
            buf.seek(0)
            msg, _, remaining = buf.read().decode('UTF-8').partition(MSG_DELIM)
            msg = msg.strip()
            self._session._dispatch_message(msg)
            self._session._buffer = StringIO()
            self._parsing_pos10 = 0
            if len(remaining.strip()) > 0:
                # There could be another entire message in the
                # buffer, so we should try to parse again.
                if type(self._session.parser) != DefaultXMLParser:
                    self.logger.debug('send remaining data to SAX parser')
                    self._session.parser.parse(remaining.encode())
                else:
                    self.logger.debug('Trying another round of parsing since there is still data')
                    self._session._buffer.write(remaining.encode())
                    self._parse10()
        else:
            # handle case that MSG_DELIM is split over two chunks
            self._parsing_pos10 = buf.tell() - MSG_DELIM_LEN
            if self._parsing_pos10 < 0:
                self._parsing_pos10 = 0

    def _parse11(self):

        """Messages are split into chunks. Chunks and messages are delimited
        by the regex #RE_NC11_DELIM defined earlier in this file. Each
        time we get called here either a chunk delimiter or an
        end-of-message delimiter should be found iff there is enough
        data. If there is not enough data, we will wait for more. If a
        delimiter is found in the wrong place, a #NetconfFramingError
        will be raised."""

        self.logger.debug("_parse11: starting")

        # suck in whole string that we have (this is what we will work on in
        # this function) and initialize a couple of useful values
        self._session._buffer.seek(0, os.SEEK_SET)
        data = self._session._buffer.getvalue()
        data_len = len(data)
        start = 0
        self.logger.debug('_parse11: working with buffer of %d bytes', data_len)
        while True and start < data_len:
            # match to see if we found at least some kind of delimiter
            self.logger.debug('_parse11: matching from %d bytes from start of buffer', start)
            re_result = RE_NC11_DELIM.match(data[start:].decode('utf-8', errors='ignore'))
            if not re_result:

                # not found any kind of delimiter just break; this should only
                # ever happen if we just have the first few characters of a
                # message such that we don't yet have a full delimiter
                self.logger.debug('_parse11: no delimiter found, buffer="%s"', data[start:].decode())
                break

            # save useful variables for reuse
            re_start = re_result.start()
            re_end = re_result.end()
            self.logger.debug('_parse11: regular expression start=%d, end=%d', re_start, re_end)

            # If the regex doesn't start at the beginning of the buffer,
            # we're in trouble, so throw an error
            if re_start != 0:
                raise NetconfFramingError('_parse11: delimiter not at start of match buffer', data[start:])

            if re_result.group(2):
                # we've found the end of the message, need to form up
                # whole message, save back remainder (if any) to buffer
                # and dispatch the message
                start += re_end
                message = ''.join(self._session._message_list)
                self._session._message_list = []
                self.logger.debug('_parse11: found end of message delimiter')
                self._session._dispatch_message(message)
                break

            elif re_result.group(1):
                # we've found a chunk delimiter, and group(1) is the digit
                # string that will tell us how many bytes past the end of
                # where it was found that we need to have available to
                # save the next chunk off
                self.logger.debug('_parse11: found chunk delimiter')
                digits = int(re_result.group(1))
                self.logger.debug('_parse11: chunk size %d bytes', digits)
                if (data_len-start) >= (re_end + digits):
                    # we have enough data for the chunk
                    fragment = textify(data[start+re_end:start+re_end+digits])
                    self._session._message_list.append(fragment)
                    start += re_end + digits
                    self.logger.debug('_parse11: appending %d bytes', digits)
                    self.logger.debug('_parse11: fragment = "%s"', fragment)
                else:
                    # we don't have enough bytes, just break out for now
                    # after updating start pointer to start of new chunk
                    start += re_start
                    self.logger.debug('_parse11: not enough data for chunk yet')
                    self.logger.debug('_parse11: setting start to %d', start)
                    break

        # Now out of the loop, need to see if we need to save back any content
        if start > 0:
            self.logger.debug(
                '_parse11: saving back rest of message after %d bytes, original size %d',
                start, data_len)
            self._session._buffer = StringIO(data[start:])
            if start < data_len:
                self.logger.debug('_parse11: still have data, may have another full message!')
                self._parse11()
        self.logger.debug('_parse11: ending')



# by Pasi Sikala <pasi.sikala@elisa.fi>

from ncclient.xml_ import *

from ncclient.operations.rpc import RPC
from ncclient.operations import util


class ShowCLI(RPC):
    def request(self, command=None):
        """Run CLI -show commands
        *command* (show) command to run

        """
        node = new_ele('get')
        filter = sub_ele(node, 'filter')
        block = sub_ele(filter, 'oper-data-format-cli-block')
        sub_ele(block, 'cli-show').text = command

        return self._request(node)


class GetConfiguration(RPC):

    def request(self, content='xml', filter=None, detail=False):
        """Get config from Alu router
        *content*   Content layer. cli or xml
        *filter*    specifies the portion of the configuration to retrieve (by default entire configuration is retrieved)
        *detail*    Show detailed config in CLI -layer"""

        node = new_ele('get-config')
        node.append(util.datastore_or_url('source', 'running', self._assert))

        if filter is not None:

            if content == 'xml':
                node.append(util.build_filter(('subtree', filter)))

            elif content == 'cli':
                rep = new_ele('filter')
                sub_filter = sub_ele(rep, 'config-format-cli-block')

                if filter is not None:
                    for item in filter:
                        if detail:
                            sub_ele(sub_filter, 'cli-info-detail').text = item
                        else:
                            sub_ele(sub_filter, 'cli-info').text = item
                else:
                    if detail:
                        sub_ele(sub_filter, 'cli-info-detail')
                    else:
                        sub_ele(sub_filter, 'cli-info')

                node.append(validated_element(rep))

        return self._request(node)


class LoadConfiguration(RPC):

    def request(self, format='xml', default_operation=None, target='running', config=None):
        node = new_ele('edit-config')
        config_node = new_ele('config')

        if default_operation is not None:
            # TODO: check if it is a valid default-operation
            sub_ele(node, "default-operation").text = default_operation

        if config is not None:
            if format == 'xml':
                node.append(util.datastore_or_url('target', target, self._assert))
                config_node.append(config)

            if format == 'cli':
                node.append(util.datastore_or_url('target', 'running', self._assert))
                sub_ele(config_node, 'config-format-cli-block').text = config

            node.append(config_node)
        return self._request(node)



import logging

class SessionLoggerAdapter(logging.LoggerAdapter):
    """Logger adapter that automatically adds session information to logs."""

    def process(self, msg, kwargs):
        if 'session' not in self.extra or self.extra['session'] is None:
            return msg, kwargs
        session = self.extra['session']
        prefix = ""
        # All Session instances have an id. SSHSessions have a host as well.
        if hasattr(session, 'host'):
            prefix += "host %s " % session.host

        if session.id is not None:
            prefix += "session-id %s" % session.id
        else:
            prefix += "session 0x%x" % id(session)

        # Pass the session information through to the LogRecord itself
        if 'extra' not in kwargs:
            kwargs['extra'] = self.extra
        else:
            kwargs['extra'].update(self.extra)

        return "[%s] %s" % (prefix, msg), kwargs
